// ===========================================================================
// 
//       Filename:  structuredgrid.cpp
// 
//    Description:  
// 
//        Version:  0.0
//        Created:  09/12/2013 08:36:55 AM
//       Revision:  none
//       Compiler:  Tested with g++ 
// 
//         Author:  M. Andy Kass (MAK)
//
//   Organisation:  US Geological Survey
//                  
//
//          Email:  mkass@usgs.gov
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ===========================================================================

#include "structuredgrid.h"

namespace formalhaut {

	std::ostream &operator<<(std::ostream &stream,
		const StructuredGrid &ob) {
		stream << "Class: StructuredGrid\n";

		return stream;
	}

	StructuredGrid::StructuredGrid() : LogObj(NULL) {

	}

	StructuredGrid::~StructuredGrid() {

	}

	void StructuredGrid::SetNCells(const Eigen::Matrix<int,3,1>& ncells) {
		this->Nx3 = ncells;
	}

	void StructuredGrid::SetOrigin(const Vector3r& origin) {
		this->Origin = origin;
	}

	void StructuredGrid::ReadMeshtoolsGrid(const std::string& fname) {

		Eigen::Matrix<int,3,1> tn;
		Vector3r to;
		VectorXr dx;
		VectorXr dy;
		VectorXr dz;

		int adds = 0;
		int multiplier;
		int bufsize;
		std::string temp;
		std::string temp2;
		std::string tast;
		char* commentBuffer = new char[200];

		bufsize = 200;
		std::fstream infile(fname.c_str(),std::ios::in);
		if (infile.fail()) {
			throw GenericFileIOError(fname);
		}
		infile.getline(commentBuffer,bufsize);
		temp = commentBuffer;
		std::istringstream in(temp);
		for (int ii=0;ii<3;ii++) {
			in >> tn(ii);
		}
		in.clear();

		infile.getline(commentBuffer,bufsize);
		temp = " ";
		temp = commentBuffer;
		std::istringstream in2(temp);
		for (int ii=0;ii<=2;ii++) {
			in2 >> to(ii);
		}
		in2.clear();

		dx.resize(tn(0));
		dy.resize(tn(1));
		dz.resize(tn(2));

		infile.getline(commentBuffer,bufsize);
		temp = commentBuffer;
		in.str(temp.c_str());
		int ii = 0;
		Real t2;
		while (ii <=tn(0)-1) {
			in >> t2;
			tast = in.peek();
			if (tast=="*") {
				temp2 = in.get();
				multiplier = t2;
				in >> t2;
				for (int jj=0;jj<=multiplier-2;jj++) {
					dx(ii+jj) = t2;
					adds = jj;
				}
				ii = ii+adds+1;
			}
			dx(ii) = t2;
			ii++;
		}

		infile.getline(commentBuffer,bufsize);
		temp = commentBuffer;
		in.str(temp.c_str());
		ii = 0;
		adds = 0;
		while (ii <=tn(1)-1) {
			in >> t2;
			tast = in.peek();
			if (tast=="*") {
				temp2 = in.get();
				multiplier = t2;
				in >> t2;
				for (int jj=0;jj<=multiplier-2;jj++) {
					dy(ii+jj) = t2;
					adds = jj;
				}
				ii = ii+adds+1;
			}
			dy(ii) = t2;
			ii++;
		}

		infile.getline(commentBuffer,bufsize);
		temp = commentBuffer;
		in.str(temp.c_str());
		ii = 0;
		adds = 0;
		while (ii <=tn(2)-1) {
			in >> t2;
			tast = in.peek();
			if (tast=="*") {
				temp2 = in.get();
				multiplier = t2;
				in >> t2;
				for (int jj=0;jj<=multiplier-2;jj++) {
					dz(ii+jj) = t2;
					adds = jj;
				}
				ii = ii+adds+1;
			}
			dz(ii) = t2;
			ii++;
		}

		

/*		
		infile.getline(commentBuffer,bufsize);
		temp = commentBuffer;
		in.str(temp.c_str());
		int ii = 0;
		while (ii <= tn(0)-1) {
			in >> dx(ii);
			std::cout << dx(ii) << std::endl;
			ii++;
			tast = in.peek();
			if (tast=="*") {
				temp2 = in.get();
				in >> multiplier;
				std::cout << multiplier << std::endl;
				for (int jj=0;jj<=multiplier-2;jj++) {
					dx(ii+jj) = dx(ii-1);
					adds = jj;
				}
				ii = ii+adds+1;
			}
		}
		in.clear();

		infile.getline(commentBuffer,bufsize);
		temp = commentBuffer;
		in.str(temp.c_str());
		ii = 0;
		adds = 0;
		while (ii <= tn(1)-1) {
			in >> dy(ii);
			ii++;
			tast = in.peek();
			if (tast == "*") {
				temp2 = in.get();
				in >> multiplier;
				for (int jj=0;jj<=multiplier-2;jj++) {
					dy(ii+jj) = dy(ii-1);
					adds = jj;
				}
				ii = ii+adds+1;
			}
		}
		in.clear();

		infile.getline(commentBuffer,bufsize);
		temp = commentBuffer;
		in.str(temp.c_str());
		ii = 0;
		adds = 0;
		while (ii <= tn(2)-1) {
			in >> dz(ii);
			ii++;
			tast = in.peek();
			if (tast=="*") {
				temp2=in.get();
				in >> multiplier;
				for (int jj=0;jj<=multiplier-2;jj++) {
					dz(ii+jj) = dz(ii-1);
					adds = jj;
				}
				ii = ii+adds+1;
			}
		}
		in.clear();
*/
		infile.close();
		delete [] commentBuffer;

		this->Origin = to;
		this->Nx3 = tn;

		if (this->TypeOfGrid == regular) {
			// Check to make sure mesh is regular
			for (int jj=0;jj<this->Nx3(0);jj++) {
				if (dx(0) != dx(jj)) {
					std::cout << "Error!  Non-regular mesh read into " <<
						"regular mesh object.  Exiting." << std::endl;
					exit(EXIT_FAILURE);
				}
			}
			for (int jj=0;jj<this->Nx3(1);jj++) {
				if (dy(0) != dy(jj)) {
					std::cout << "Error!  Non-regular mesh read into " <<
						"regular mesh object.  Exiting." << std::endl;
					exit(EXIT_FAILURE);
				}
			}
			for (int jj=0;jj<this->Nx3(2);jj++) {
				if (dz(0) != dz(jj)) {
					std::cout << "Error!  Non-regular mesh read into " <<
						"regular mesh object.  Exiting." << std::endl;
					exit(EXIT_FAILURE);
				}
			}

			// Resize Dx3 and populate
			this->Dx3.resize(Eigen::NoChange,1);
			this->Dx3(0) = dx(0);
			this->Dx3(1) = dy(0);
			this->Dx3(2) = dz(0);

			// Write to log file
			std::string logw;
			std::stringstream ss;
			std::vector<std::string> logs;
			this->LogObj->WriteLogLine("Regular Mesh parameters:");
			logw = "nx\t\tny\t\tnz\t\tdx\t\tdy\t\tdz\t\tox\t\toy\t\toz";
			this->LogObj->WriteLogLine(logw);
			this->LogObj->WriteDashedLine();
			ss << this->Nx3(0);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Nx3(1);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Nx3(2);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Dx3(0);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Dx3(1);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Dx3(2);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Origin(0);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Origin(1);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Origin(2);
			logs.push_back(ss.str());
			ss.str("");
			logw = "";
			for (int jj=0;jj<9;jj++) {
				logw = logw + logs[jj] + "\t\t";
			}
			this->LogObj->WriteLogLine(logw);
			this->LogObj->WriteLogLine("\n");
			

		}
		else if (this->TypeOfGrid == rectilinear) {
			this->Dx3v.push_back(dx);
			this->Dx3v.push_back(dy);
			this->Dx3v.push_back(dz);

			// Write to log file
			std::string logw;
			std::stringstream ss;
			std::vector<std::string> logs;
			this->LogObj->WriteLogLine("Rectilinear Mesh Parameters:");
			logw = "nx\t\tny\t\tnz\t\tox\t\toy\t\toz";
			this->LogObj->WriteLogLine(logw);
			this->LogObj->WriteDashedLine();
			ss << this->Nx3(0);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Nx3(1);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Nx3(2);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Origin(0);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Origin(1);
			logs.push_back(ss.str());
			ss.str("");
			ss << this->Origin(2);
			logs.push_back(ss.str());
			ss.str("");
			logw = "";
			for (int jj=0;jj<6;jj++) {
				logw = logw + logs[jj] + "\t\t";
			}
			this->LogObj->WriteLogLine(logw);
			this->LogObj->WriteLogLine("\n");


		}
		else if (this->TypeOfGrid == octree) {
			std::cout << "Error!  Octree mesh object cannot interface " <<
				"with Meshtools format.  Exiting." << std::endl;
			exit(EXIT_FAILURE);
		}
		else {
			std::cout << "Error!  Mesh object not declared properly. Exiting."
				<< std::endl << "(will try to be more unhelpful in the "
				<< "future)" << std::endl;
			exit(EXIT_FAILURE);
		}


	}

	void StructuredGrid::AttachLogObject(Logging* logobj) {
		this->LogObj = logobj;
	}

	int StructuredGrid::GetVectorPosition(const Eigen::Matrix<int,3,1>& 
		cellpos) {

		int vecpos;

		// Check and make sure the points are valid
		if (cellpos(0) >= this->Nx3(0)) {
			std::cout << "Invalid queried vector position in x dim\n";
			exit(EXIT_FAILURE);
		}
		if (cellpos(1) >= this->Nx3(1)) {
			std::cout << "Invalid queried vector position in y dim\n";
			exit(EXIT_FAILURE);
		}
		if (cellpos(2) >= this->Nx3(2)) {
			std::cout << "Invalid queried vector position in z dim\n";
			exit(EXIT_FAILURE);
		}

		// z changes first, then x, then y
		vecpos = (cellpos(1)*this->Nx3(0)*this->Nx3(2)) +
			(cellpos(0)*this->Nx3(2)) + cellpos(2);

		return vecpos;
	}

	int StructuredGrid::QueryCellPosition(const Vector3r& cellpos) {
		int vecpos;
		Vector3r cellposmo;

		//Check to make sure the point is in the model region
		if (cellpos(0) < this->Origin(0) || cellpos(0) >= 
			this->Nx3(0)*this->Dx3(0) + this->Origin(0)) {
			std::cout << "Queried point exists outside model region.\n";
			std::cout << "Given location: " << cellpos(0) << "\n"
				<< "Region limits: " << this->Origin(0) << " to " <<
				this->Nx3(0)*this->Dx3(0)+this->Origin(0) << std::endl;
			exit(EXIT_FAILURE);
		}
		
		if (cellpos(1) < this->Origin(1) || cellpos(1) >= 
			this->Nx3(1)*this->Dx3(1) + this->Origin(1)) {
			std::cout << "Queried point exists outside model region.\n";
			std::cout << "Given location: " << cellpos(1) << "\n"
				<< "Region limits: " << this->Origin(1) << " to " <<
				this->Nx3(1)*this->Dx3(1)+this->Origin(1) << std::endl;
			exit(EXIT_FAILURE);
		}

		if (cellpos(2) < this->Origin(2) || cellpos(2) >= 
			this->Nx3(2)*this->Dx3(2) + this->Origin(2)) {
			std::cout << "Queried point exists outside model region.\n";
			std::cout << "Given location: " << cellpos(2) << "\n"
				<< "Region limits: " << this->Origin(2) << " to " <<
				this->Nx3(2)*this->Dx3(2)+this->Origin(2) << std::endl;
			exit(EXIT_FAILURE);
		}

		//Subtract off the origin just to make the maths easier
		for (int ii=0;ii<3;ii++) {
			cellposmo(ii) = cellpos(ii) - this->Origin(ii);
		}

		vecpos = 42;
		/// \todo Finish this up.  Returning 42 right now.
		return vecpos;		
	}

	Eigen::Matrix<int,3,1> StructuredGrid::GetNCells() {
		return this->Nx3;
	}

	Vector3r StructuredGrid::GetOrigin() {
		return this->Origin;
	}

	Vector3Xr StructuredGrid::GetCentroids() {
		return this->VecCentroid;
	}

	Vector3r StructuredGrid::GetCentroid(const int& vecpos) {
		return this->VecCentroid.row(vecpos);
	}












}
